from math import *
import string
import re
import os.path
import datetime
import traceback
import time

import win32com.client
from pythoncom import CoInitialize, CoUninitialize

import qubtree
from Buttons import AddTaskButton
from Dialogs import *
import QUB
try:
    from qubopt import BalanceAllLoops, PareConstraints
except:
    def BalanceAllLoops(mdl):
        print "HyperQub requires a 2006 QUB for detailed balance constraints"
        return mdl
    def PareConstraints(mdl):
        print "HyperQub requires a 2006 QUB to eliminate duplicate constraints"

def concat(lsts):
    return reduce(lambda x,y:x+y, lsts)

def cartprod(*lsts):
    def extend(prod, remains):
        if 0 == len(remains):
            return [prod]
        subremains = remains[1:]
        return concat([extend(prod+[el], subremains) for el in remains[0]])
    return extend([], lsts)

def EvensOdds(lst):
    evens = [x for i, x in enumerate(lst) if not (i%2)]
    odds  = [x for i, x in enumerate(lst) if (i%2)]
    return evens, odds

class Anon:
    def __init__(self, **kw):
        for field in kw.keys():
            self.__dict__[field] = kw[field]
    def __repr__(self):
        return str(self.__dict__)
    def __str__(self):
        return str(self.__dict__)

class Vertex(Anon):
    def __init__(self, **kw):
        self.lbl = ''
        self.color = 0
        self.pr = 0.0
        self.x = 0.0
        self.y = 0.0
        Anon.__init__(self, **kw)
        self.outs = []
    # i, lbl, color, pr, x, y, subvertices, outs
class Edge(Anon):
    def __repr__(self):
        return self.__str__()
    def __str__(self):
        try:
            return '%s -> %s by %s' % (self.v.lbl, self.w.lbl, self.lbl)
        except:
            try:
                return '%s -> %s' % (self.v.lbl, self.w.lbl)
            except:
                return '%d -> %d' % (self.v.i, self.w.i)
         # v, w, rate,
         # lbl e.g. '+C--r',
         # ngrp (same neighbor conditions) e.g. '+C-r',
         # dpgrp (dimension and on-or-off-path) e.g. 'Conr'
         # subedges
class Graph:
    def __init__(self):
        self.vertices = []
        self.constraints = []
    def RevEdge(self, e):
        for r in e.w.outs:
            if r.w == e.v:
                return r
    def FindEdge(self, a, b):
        for i, e in enumerate(self.vertices[a].outs):
            if e.w.i == b:
                return e
        e = Edge(v=self.vertices[a], w=self.vertices[b], lbl='%d->%d'%(a,b))
        self.vertices[a].outs.append(e)
        return e
    def ToTree(self, baseMdl = None, addMissing=False, readColors=False, remExtra=False):
        if baseMdl != None:
            self.ReadVertices(baseMdl, addMissing=addMissing, readColors=readColors, remExtra=remExtra) # keep the user's positions and colors
            # self.ReadEdges(baseMdl)
            # self.ReadConstraints(baseMdl)
            mdl = baseMdl
        else:
            mdl = qubtree.Node('ModelFile')
            mdl['Amps'].data = map(float, range(10))
            mdl['Stds'].data = [0.1] * 10
            mdl['NAr'].data = [0] * 10
        self.WriteVertices(mdl)
        self.WriteEdges(mdl)
        self.WriteConstraints(mdl)
        return mdl
    def FromTree(self, mdl):
        self.ReadVertices(mdl)
        self.ReadEdges(mdl)
        self.ReadConstraints(mdl)
    def ReadVertices(self, mdl, addMissing=True, readColors=True, remExtra=True):
        state = mdl.find('States').find('State')
        i = 0
        while not state.isNull:
            if i >= len(self.vertices):
                if addMissing:
                    self.vertices.append(Vertex(i=len(self.vertices)))
                else:
                    break
            v = self.vertices[i]
            if readColors:
                v.color = state['Class'].data[0]
            v.pr = state['Pr'].data[0]
            v.x = state['x'].data[0] or v.x
            v.y = state['y'].data[0] or v.y
            state = state.nextSameName()
            i += 1
        if remExtra:
            self.vertices[i:] = []
    def WriteVertices(self, mdl):
        states = mdl['States']
        state = None
        for v in self.vertices:
            if state == None:
                state = states['State']
            elif state.nextSameName().isNull:
                state = states.insert('State', state)
            else:
                state = state.sibling
            state['Class'].data = v.color
            state['Pr'].data = v.pr
            state['x'].data = v.x
            state['y'].data = v.y
    def ReadEdges(self, mdl):
        stillExists = {}
        rate = mdl.find('Rates').find('Rate')
        while not rate.isNull:
            a, b = rate['States'].data
            k0 = rate['k0'].data
            for i, e in enumerate([self.FindEdge(a,b), self.FindEdge(b,a)]):
                e.rate = k0[i]
                stillExists[e] = True
            rate = rate.nextSameName()
        for v in self.vertices:
            keep = []
            for e in v.outs:
                if stillExists.has_key(e):
                    keep.append(e)
            v.outs[:] = keep
    def WriteEdges(self, mdl):
        rates = mdl['Rates']
        while not rates.child.isNull:
            rates.remove(rates.child)
        rate = qubtree.NullNode()
        for v in self.vertices:
            for e in v.outs:
                if e.v.i < e.w.i:
                    e2 = self.FindEdge(e.w.i, e.v.i)
                    rate = rates.insert('Rate', rate)
                    rate['States'].data = (int(e.v.i), int(e.w.i))
                    rate['k0'].data = (e.rate, e2.rate)
                    rate['k1'].data = (0.0, 0.0)
                    rate['P'].data = (0, 0)
                    rate['Q'].data = (0, 0)
    def ReadConstraints(self, mdl):
        self.constraints[:] = []
        cns = mdl['Constraints'].child
        while not cns.isNull:
            self.constraints.append( (cns.name, cns.data[:]) )
            cns = cns.sibling
    def WriteConstraints(self, mdl):
        cnss = mdl['Constraints']
        while not cnss.child.isNull:
            cnss.remove(cnss.child)
        cns = qubtree.NullNode()
        for name, states in self.constraints:
            cns = cnss.insert(name, cns)
            cns.data = states
    def CircleLayout(self):
        n = len(self.vertices)
        for i, v in enumerate(self.vertices):
            theta = pi/2 + i*2*pi/n
            v.x = 50.0 + 40.0*cos(theta)
            v.y = 50.0 - 40.0*sin(theta)

def LinRates(graph):
    trace = []
    visited = {}
    def Search(v):
        visited[v] = True
        if len(v.outs) > 2:
            return []
        for e in v.outs:
            if not visited.has_key(e.w):
                return [v] + Search(e.w)
        return [v]
    xmin = 101.0
    vstart = None
    for v in graph.vertices:
        if len(v.outs) == 1 and v.x < xmin:
            xmin = v.x
            vstart = v
    if vstart:
        trace = Search(vstart)
    def BiRate(v, w):
        e = graph.FindEdge(v.i, w.i)
        return (e.rate, graph.RevEdge(e).rate)
    return [BiRate(trace[i], trace[i+1]) for i in range(len(trace)-1)]

class RateSheet:
    def __init__(self, linRates, defOffPath):
        #CoInitialize() -- let WordReport.py do this on startup

        try:
            win32com.client.gencache.EnsureModule('{00020813-0000-0000-C000-000000000046}',0,1,3)
        except:
            pass

        self.excel = win32com.client.Dispatch("Excel.Application")
        self.book = self.excel.Workbooks.Add(-4167) # XLWBatWorksheet
        self.sheet = self.book.Worksheets.Item(1)
        cells = self.sheet.Cells
        self.excel.Visible = 1

        n = len(linRates)
        self.n = n
        dims = [chr(ord('A')+i) for i in range(n)]
        
        def WriteRow(r, c, *args):
            for i, arg in enumerate(args):
                cells(r, c+i).Value = arg
        rk0h = 1
        ck0h = 1
        WriteRow(rk0h, ck0h, '', 'k0 fwd', 'k0 bwd', 'fix (blank for no)')
        rk0on = rk0h + 1
        for i in range(n):
            WriteRow(rk0on+i, ck0h, dims[i] + 'on', linRates[i][0], linRates[i][1], 'X')
        rk0off = rk0on + n + 1
        r = rk0off
        def EnumGroups(i):
            syms = []
            if i > 0:
                syms.append(['-','+'])
            if i > 1:
                syms.append(['-','+'])
            syms.append([dims[i]])
            if (i+1) < n:
                syms.append(['-','+'])
            if (i+2) < n:
                syms.append(['-','+'])
            for pieces in cartprod(*syms):
                word = string.join(pieces,'')
                if not re.match('\\+*[A-Z]-*$', word):
                    yield word
        for i in range(n):
            for group in EnumGroups(i):
                WriteRow(r, ck0h, group, defOffPath, defOffPath, '')
                r += 1
            # old 1-neighbor style
            #if i == 0:
            #    WriteRow(r, ck0h, dims[i] + '-+', defOffPath, defOffPath, '')
            #    r += 1
            #    WriteRow(r, ck0h, dims[i] + '+-', defOffPath, defOffPath, '')
            #    r += 1
            #    WriteRow(r, ck0h, dims[i] + '++', defOffPath, defOffPath, '')
            #    r += 1
            #elif i == (n-1):
            #    WriteRow(r, ck0h, '--' + dims[i], defOffPath, defOffPath, '')
            #    r += 1
            #    WriteRow(r, ck0h, '-+' + dims[i], defOffPath, defOffPath, '')
            #    r += 1
            #    WriteRow(r, ck0h, '+-' + dims[i], defOffPath, defOffPath, '')
            #    r += 1
            #else:
            #    WriteRow(r  , ck0h, '-'+dims[i]+'-', defOffPath, defOffPath, '')
            #    WriteRow(r+1, ck0h, '-'+dims[i]+'+', defOffPath, defOffPath, '')
            #    WriteRow(r+2, ck0h, '+'+dims[i]+'+', defOffPath, defOffPath, '')
            #    r += 3
        cells(r+2, 1).Value = 'To remove connections, set fwd and bwd to 0'
        self.rclear = r+4
        self.rk0h = rk0h
        self.ck0h = ck0h
        self.rk0on = rk0on
        self.rk0off = rk0off
    #def __del__(self):
        #self.excel.DisplayAlerts = 0
        #self.book.Close()
        #self.excel.DisplayAlerts = 1
        #CoUninitialize()
    def GetRates(self):
        rates = {}
        def GatherRates(r):
            cells = self.sheet.Cells
            while cells(r, self.ck0h).Value:
                lbl = cells(r, self.ck0h).Value
                rates[lbl] = float(cells(r, self.ck0h+1).Value)
                rates[lbl+'r'] = float(cells(r, self.ck0h+2).Value)
                rates[lbl+'fix'] = bool(cells(r, self.ck0h+3).Value)
                rates[lbl+'rfix'] = bool(cells(r, self.ck0h+3).Value)
                r += 1
        GatherRates(self.rk0on)
        GatherRates(self.rk0off)
        return rates

class Hypercube(Graph):
    def __init__(self, n):
        Graph.__init__(self)
        self.n = n
        # vertices
        dims = [chr(ord('A')+i) for i in range(n)]
        syms = [ ['-', dim] for dim in dims ]
        for state in apply(cartprod, syms):
            self.vertices.append(Vertex(i=len(self.vertices), lbl=string.join(state,'')))
        self.vertices[-1].color = 1
    def SetupRates(self, rates={'offpath':0.1}, scale='neighbors', balance=True):#scale='onoff' or 'none'
        self.rates = rates
        n = self.n
        dims = [chr(ord('A')+i) for i in range(n)]
        bits = [int(round(pow(2, i))) for i in range(n)]
        bits.reverse()                  # [..., 4, 2, 1]
        def Labels(s0, s1):
            have = [ [bool(s & bit) for s in [s0, s1]] for bit in bits ]
            key  = {True:'+', False:'-'}
            lbl  = string.join([key[hav[0]] for hav in have], '')
            for i, hav in enumerate(have):
                if hav[0] != hav[1]:
                    dim = i
                    lbl = lbl[:i] + dims[i] + lbl[i+1:]
                    break
            ngrp = lbl[max(0,dim-2) : min(n,dim+3)] # (old 1-neighbor style:) lbl[max(0,dim-1) : min(n,dim+2)]
            dpgrp = dims[dim]
            if (ngrp == 'A--') or (ngrp == '++'+dims[n-1]) or re.match('\\+B--', ngrp) or re.search('\\+\\+[A-Z]--', ngrp) or re.search('\\+\\+[A-Z]-$', ngrp):
            # old 1-neighbor style:
            # if (ngrp == 'A--') or (ngrp == '++'+dims[n-1]) or (len(ngrp)>=3 and ngrp[0] == '+' and ngrp[2] == '-'):
                dpgrp += 'on'
            else:
                dpgrp += 'off'
            if have[dim][0]:
                lbl += 'r'
                ngrp += 'r'
                dpgrp += 'r'
            return lbl, ngrp, dpgrp
        def GetRate(edge):
            try:
                return rates[edge.ngrp]
            except:
                try:
                    return rates[edge.dpgrp]
                except:
                    return rates['offpath']
        for v in self.vertices:
            v.outs[:] = [] # clear edges
        marked = {}
        for vi in range(len(self.vertices)):
            marked[vi] = False
        def AddEdgesFrom(vi):
            if not marked[vi]:
                marked[vi] = True
                for i, b in enumerate(bits):
                    if not (vi & b):
                        for x, y in [(vi, vi+b), (vi+b, vi)]:
                            lbl, ngrp, dpgrp = Labels(x, y)
                            edge = Edge(v = self.vertices[x],
                                        w = self.vertices[y],
                                        lbl=lbl, ngrp=ngrp, dpgrp=dpgrp)
                            edge.rate = GetRate(edge)
                            if edge.rate > 0.0:
                                self.vertices[x].outs.append(edge)
                        AddEdgesFrom(vi+b)
        AddEdgesFrom(0)
        #constraints
        self.constraints[:] = []
        def IsFixed(e):
            if rates.has_key(e.ngrp+'fix'):
                return rates[e.ngrp+'fix']
            if rates.has_key(e.dpgrp+'fix'):
                return rates[e.dpgrp+'fix']
            return False
        for v in self.vertices:
            for e in v.outs:
                if IsFixed(e):
                    self.constraints.append(('FixRate', (e.v.i, e.w.i)))
        if scale == 'neighbors':
            GroupOf = lambda e: e.ngrp
        elif scale == 'onoff':
            GroupOf = lambda e: e.dpgrp
        if scale != 'none':
            scaleGroups = {}
            def AddToGroup(e):
                group = GroupOf(e)
                if scaleGroups.has_key(group):
                    scaleGroups[group].append(e)
                else:
                    scaleGroups[group] = [e]
            for v in self.vertices:
                for e in v.outs:
                    AddToGroup(e)
            for group in scaleGroups.values():
                if len(group) > 1:
                    a = group[0].v.i
                    b = group[0].w.i
                    for i in range(1, len(group)):
                        self.constraints.append(('ScaleRate', (a, b, group[i].v.i, group[i].w.i)))
            mdl = self.ToTree()
            mdl = PareConstraints(mdl)
        if balance:
            mdl = BalanceAllLoops(mdl)
            self.ReadConstraints(mdl)
    def Layout(self):
        def Place(v, x, y):
            v.x = x
            v.y = y
        def PlaceLine(vixs, x, y, w, h):
            ymid = y + h/2
            Place(self.vertices[vixs[0]], x, ymid)
            Place(self.vertices[vixs[1]], x+w, ymid)
        def PlaceSquare(vixs, x, y, w, h):
            evens, odds = EvensOdds(vixs)
            PlaceLine(evens, x, y+h, w, 0.0)
            PlaceLine(odds,  x, y,   w, 0.0)
        def PlaceCube(vixs, x, y, w, h):
            evens, odds = EvensOdds(vixs)
            xsub = w/5
            ysub = h/5
            PlaceSquare(evens, x, y+ysub, w-xsub, h-ysub)
            PlaceSquare(odds, x+ysub, y, w-xsub, h-ysub)
        def PlaceTesseract(vixs, x, y, w, h):
            evens, odds = EvensOdds(vixs)
            PlaceCube(evens, x+w/3, y+h/3, w/3, h/3)
            PlaceCube(odds, x, y, w, h)
        x, y, w, h = 10.0, 15.0, 80.0, 75.0
        if self.n == 1:
            PlaceLine(range(2), x, y, w, h)
        elif self.n == 2:
            PlaceSquare(range(4), x, y, w, h)
        elif self.n == 3:
            PlaceCube(range(8), x, y, w, h)
        elif self.n == 4:
            PlaceTesseract(range(16), x, y, w, h)
        else:
            self.CircleLayout()


def CalcPeq(graph, lbl, scratch):
    graph.CircleLayout()
    scratch.setAsNode(graph.ToTree())
    scratch.CalcPeq()
    graph.ReadVertices(scratch.getAsNode())
    prMin = 1.0e10
    prSum = 0.0
    for v in graph.vertices:
        prMin = min(v.pr, prMin)
        prSum += v.pr
    if prMin < 0.0 or abs(prSum - 1.0) > 0.01:
        QUB.Report('Error calculating Peq for group %s: setting all equal' % lbl)
        avg = 1.0 / len(graph.vertices)
        for v in graph.vertices:
            v.pr = avg
    # ok until CalcPeqAndAssign is there just make them even:
    #avg = 1.0 / len(graph.vertices)
    #for v in graph.vertices:
    #    v.pr = avg

class HyperGrouped(Graph):
    def __init__(self, hq, isfast, scratch):
        Graph.__init__(self)
        self.hq = hq
        #groups
        marked = {}
        for v in hq.vertices:
            marked[v] = False
        def GroupOf(v):
            group = [v]
            marked[v] = True
            for edge in v.outs:
                if (not marked[edge.w]) and (v.color == edge.w.color) and isfast(edge.rate) and isfast(hq.RevEdge(edge).rate):
                    group += GroupOf(edge.w)
            return group
        def GroupPeq(group, lbl):
            groupMdl = Graph()
            fullToSub = {}
            for i, v in enumerate(group):
                fullToSub[v.i] = i
            for i, v in enumerate(group):
                vsub = Vertex(i=i, lbl=v.lbl, fulli=v.i)
                groupMdl.vertices.append(vsub)
            for i, v in enumerate(group):
                vsub = groupMdl.vertices[i]
                for e in v.outs:
                    if e.w in group:
                        vsub.outs.append(Edge(v=vsub, w=groupMdl.vertices[fullToSub[e.w.i]], rate=e.rate))
            CalcPeq(groupMdl, lbl, scratch)
            for full, sub in fullToSub.items():
                hq.vertices[full].pr = groupMdl.vertices[sub].pr
        # vertex per group
        for v in hq.vertices:
            if not marked[v]:
                group = GroupOf(v)
                GroupPeq(group, str(len(self.vertices)+1))
                # print 'group: ', group
                self.vertices.append(Vertex(i=len(self.vertices), subvertices=group, color=group[0].color))
        self.CircleLayout()
        # edges
        def GetRate(subedges):
            return sum([e.v.pr * e.rate for e in subedges])
        def GetRateFormula(subedges):
            formula = ''
            for i, e in enumerate(subedges):
                if i:
                    formula += '+ '
                formula += '%f*[%s]' % (e.v.pr, e.ngrp)
            return formula
        for a, b in cartprod(range(len(self.vertices)), range(len(self.vertices))):
            if a < b: #0, 1; 0, 2; 0, 3; ...; 1, 2; 1, 3; ...
                x, y = self.vertices[a], self.vertices[b]
                #print '(%d, %d)' % (a, b)
                subedges, subrevs = [], []
                for v in x.subvertices:
                    if v != x:
                        #print '  %s' % v.lbl
                        for e in v.outs:
                            if e.w in y.subvertices:
                                #print '\t%s\t%s' % (e.lbl, e.w.lbl)
                                subedges.append(e)
                                subrevs.append(hq.RevEdge(e))
                if subedges:
                    edge = Edge(v=x, w=y, subedges=subedges, rate=GetRate(subedges), rateFormula=GetRateFormula(subedges))
                    x.outs.append(edge)
                    edge = Edge(v=y, w=x, subedges=subrevs, rate=GetRate(subrevs), rateFormula=GetRateFormula(subrevs))
                    y.outs.append(edge)


def GroupReport(group, sheet, r0, c0):
    cells = sheet.Cells
    r = r0
    for v in group.vertices:
        cells(r, c0).Value = v.i + 1
        for i, sub in enumerate(v.subvertices):
            cells(r, c0+i+1).Value = '%s:%f' % (sub.lbl, sub.pr)
        r += 1
        for out in v.outs:
            cells(r, c0).Value = ''
            cells(r, c0+1).Value = out.w.i + 1
            cells(r, c0+2).Value = out.rate
            for i, sub in enumerate(out.subedges):
                cells(r, c0+i+3).Value = '%f[%s]' % (sub.v.pr, sub.ngrp)
            r += 1
    for i in range(10):
        for j in range(10):
            cells(r+i, c0+j).Value = ''

#rates = {'Conr': 10000, 'Donr': 80, 'Don': 10000, 'Aonr': 10000, 'Aon': 80, 'Eon': 10000, 'offpath': 0.10000000000000001, 'Bon': 10000, 'Bonr': 10000, 'Eonr': 10000, 'Con': 10000}
#
#hq = Hypercube(4, rates)
#hg = HyperGrouped(hq,lambda x:x>1)
#
#for v in hg.vertices:
#    for e in v.outs:
#        print e,'\t',e.getRate()
#
#for v in hg.vertices:
#    print v.i, ' ', 
#    for s in v.subvertices:
#        print s.lbl, ' ',
#    print

def FindOrNewModel(name):
    for mdl in QUB.ModelView:
        if name == str(mdl.getAsNode()['FileName'].data):
            return mdl
    QUB.ModelView.New(name)
    return QUB.ModelView[len(QUB.ModelView)-1]

def ClearTree(tree):
    while not tree.child.isNull:
        tree.remove(tree.child)

def ClearModel(mdl):
    tree = mdl.getAsNode()
    ClearTree(tree['States'])
    ClearTree(tree['Rates'])
    mdl.setAsNode(tree)

class HyperQuber:
    def __init__(self):
        frontMdl = QUB.ModelView.currModel
        self.linear = None
        self.rateSheet = None
        self.scratch = FindOrNewModel('HQ scratch.qmf')
        self.hqModel = FindOrNewModel('HyperQub.qmf')
        self.hqGraph = None
        self.groupModel = FindOrNewModel('HyperGrouped.qmf')
        self.groupGraph = None
        QUB.ModelView.Choose(frontMdl)
        
        self.wiz = QUB.CreateWizard()
        self.wiz.caption = 'HyperQub'
        self.wiz.buttons = ['Read lin', 'Sheet->HQ', 'HQ->Groups', 'Done']
        self.wizActions = {'Read lin'   : self.ReadLin,
                           'Sheet->HQ'  : self.SheetToHQ,
                           'HQ->Groups' : self.HQToGroups}
        self.message = 'Ready'
        self.wiz.message = self.message

        self.properties = QUB.Profiles['HyperQub'].active
        def initProp(name, val):
            if self.properties.find(name).isNull:
                self.properties[name].data = val
        initProp('defOff', 0.1)
        initProp('fastThresh', 1.0)
        initProp('scale', 'neighbors')
        initProp('balLoops', 1)
    def ReadButtons(self):
        self.wiz.show()
        btn = None
        while btn != 'Done':
            btn = self.wiz.waitForButton()
            if btn != 'Done':
                self.wiz.message = btn+'...'
                try:
                    self.wizActions[btn]()
                except:
                    traceback.print_exc()
            self.wiz.message = self.message
        self.wiz.hide()
    def ReadLin(self):
        self.linear = QUB.ModelView.currModel
        if not self.linear:
            QUB.Report('HQ: no model')
            return
        linGraph = Graph()
        linGraph.FromTree(self.linear.getAsNode())
        linRates = LinRates(linGraph)
        if not linRates:
            QUB.Report('HQ: not a linear model')
            return
        props = self.properties
        response = QUB.Dialog('HQ: Read linear model', ['OK', 'Cancel'],
                              [DialogText('defOff', 'Default off-path rate:', props['defOff'].data[0], acceptFloatGreaterThan(0.0))])
        if response['button'] == 'OK':
            self.rateSheet = RateSheet(linRates, response['defOff'])
            self.hqGraph = Hypercube(len(linRates))
            self.hqGraph.Layout()
            self.hqModel.setAsNode(self.hqGraph.ToTree())
            ClearModel(self.groupModel)
            self.message = 'Set up rate sheet.\r\nReady'
        props['defOff'].data = response['defOff']
    def SheetToHQ(self):
        if not self.rateSheet:
            return
        props = self.properties
        response = QUB.Dialog('HQ: Sheet to Hypercube', ['OK', 'Cancel'],
                              [DialogRadioGroup('scale', 'Scale groups:', ['neighbors', 'onoff', 'none'], str(props['scale'].data)),
                               DialogCheckBox('balLoops', 'Detailed balance', props['balLoops'].data[0])])
        if response['button'] == 'OK':
            rates = self.rateSheet.GetRates()
            self.hqGraph.SetupRates(rates, response['scale'], response['balLoops'])
            self.hqModel.setAsNode(self.hqGraph.ToTree(self.hqModel.getAsNode()))
            QUB.ModelView.Choose(self.hqModel)
            self.message = 'Set up Hypercube.\r\nReady'
            # make hqModel visible
        props['scale'].data = str(response['scale'])
        props['balLoops'].data = int(response['balLoops'])
    def HQToGroups(self):
        props = self.properties
        response = QUB.Dialog('HQ: Hypercube to Groups', ['OK', 'Cancel'],
                              [DialogText('fastThresh', '"fast" threshold:', props['fastThresh'].data[0], acceptFloatGreaterThan(0.0))])
        if response['button'] == 'OK':
            self.groupGraph = HyperGrouped(self.hqGraph, lambda x: x>response['fastThresh'], self.scratch)
            self.groupModel.setAsNode(self.groupGraph.ToTree(self.groupModel.getAsNode()))
            QUB.ModelView.Choose(self.groupModel)
            GroupReport(self.groupGraph, self.rateSheet.sheet, self.rateSheet.rclear, self.rateSheet.ck0h)
            self.message = 'Set up group model.\r\nReady'
            #make groupModel visible
        props['fastThresh'].data = response['fastThresh']


def HyperQub(props):
    try:
        hq = QUB.hq
    except:
        hq = HyperQuber()
        QUB.hq = hq
    try:
        hq.ReadButtons()
    except:
        del QUB.hq
        #traceback.print_exc()


AddTaskButton( 'HyperQub', # button text
               HyperQub,   # function
               'HQ',       # short text for small buttons
               0,            # show settings dialog on left-click?
               0,            # show settings dialog on right-click?
               "Hypercube model building and visualization",   # tool-tip
               "Hypercube model building and visualization",   # description in "Edit Buttons"
               "Properties:", # caption for the settings dialog
               [],
               'Modeling')



